-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("achievementTracker", "AceEvent-3.0", "AceHook-3.0")
local _, _, _, interfaceVersion = GetBuildInfo()

-----------------------
--- Library loading ---
-----------------------
local LATr = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-AchievementTracker")
local LSM = LibStub("LibSharedMedia-3.0")
local LibQTip = LibStub("LibQTip-1.0")

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------


-----------------
--- Variables ---
-----------------
module.modName = LATr["Achievement Tracker"]
local prdb, chdb, options
local numTrackedAchievement = 0
local achievements = {}
local achievementsByCategory = {}
local tracker, trackerAnchor

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		point = "CENTER",
		x = 0,
		y = 0,
		showAnchor = true,
		anchorPosition = 2,
		toggleTracker = true,
		hideTrackerInCombat = true,
		font = "Friz Quadrata TT",
		headerFontSize = 12,
		normalFontSize = 10,
		maxWidth = 300,
		objectivescellpadding = 10,
		scale = 1,
		backDropColor = {0.09, 0.09, 0.09, 1},
		borderColor = {1, 1, 1, 1},
	},
}

----------------------
--- Initialization ---
----------------------
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("achievementTracker", defaults, "Default")
	prdb = self.db.profile
	chdb = self.db.char
	chdb.achievementTracked = chdb.achievementTracked or {}
	chdb.categoryDisplay = chdb.categoryDisplay or {}
end

----------------------------
--- Tracker Anchor Frame ---
----------------------------
local anchor = CreateFrame("Button", nil, UIParent)
anchor:SetHeight(24)
anchor:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = {left = 5, right = 5, top = 5, bottom = 5}, tile = true, tileSize = 16})
anchor:SetBackdropColor(0.09, 0.09, 0.19, 0.5)
anchor:SetBackdropBorderColor(0.5, 0.5, 0.5, 1)
local text = anchor:CreateFontString(nil, nil, "GameFontNormalSmall")
text:SetPoint("CENTER")
text:SetText(LATr["Achievement Tracker"])
anchor:SetWidth(text:GetStringWidth() + 8)
anchor:SetMovable(true)
anchor:RegisterForDrag("LeftButton")
anchor:SetClampedToScreen(true)
anchor:SetScript("OnDragStart", function(self)
	self:StartMoving()
end)
anchor:SetScript("OnDragStop", function(self)
	self:StopMovingOrSizing()
	prdb.point, prdb.x, prdb.y = "BOTTOMLEFT", self:GetLeft(), self:GetBottom()
	trackerAnchor = anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	LibQTip:Release(tracker)
	module:updateAchievementTracker()
end)

---------------
--- Options ---
---------------
function module:getOptions()
	if not options then
		options = {
			type = "group",
			name = module.modName,
			desc = module.modName,
			get = function(info) 
				if type(prdb[info[#info]]) == "table" then
					return unpack(prdb[info[#info]])
				else
					return prdb[info[#info]]
				end
			end,
			set = function(info, arg1, arg2, arg3, arg4)
				if arg2 then
					prdb[info[#info]][1] = arg1
					prdb[info[#info]][2] = arg2
					prdb[info[#info]][3] = arg3
					prdb[info[#info]][4] = arg4
				else
					prdb[info[#info]] = arg1
				end
				if info[#info] == "showAnchor" then
					if prdb.showAnchor then
						anchor:Show()
					else
						anchor:Hide()
					end
				elseif info[#info] == "hideTrackerInCombat" then
					if prdb.hideTrackerInCombat then
						module:RegisterEvent("PLAYER_REGEN_ENABLED","updateQuestsTracker")
						module:RegisterEvent("PLAYER_REGEN_DISABLED","updateQuestsTracker")
					else
						module:UnregisterEvent("PLAYER_REGEN_ENABLED")
						module:UnregisterEvent("PLAYER_REGEN_DISABLED")
					end
				end
				LibQTip:Release(tracker)
				module:updateAchievementTracker()
			end,
			args = {
				showAnchor = {
					name = LATr["Show quest tracker anchor"],
					type = "toggle",
					width = "full",
					order = 6,
				},
				anchorPosition = {
					name = LATr["Anchor position"],
					type = "select",
					values = {[1] = "TOPLEFT", [2] = "TOP", [3] = "TOPRIGHT", [4] = "BOTTOMLEFT", [5] = "BOTTOM", [6] = "BOTTOMRIGHT"},
					width = "full",
					order = 7,
				},
				toggleTracker = {
					name = LATr["Display tracker"],
					type = "toggle",
					width = "full",
					order = 9,
				},
				hideTrackerInCombat = {
					name = LATr["Hide quest tracker in combat"],
					type = "toggle",
					width = "full",
					order = 10,
				},
				font = {
					name = LATr["Font"],
					type = "select",
					dialogControl = 'LSM30_Font',
					values = AceGUIWidgetLSMlists.font,
					width = "full",
					order = 13,
				},
				headerFontSize = {
					name = LATr["Achievement font size"],
					type="range",
					min = 6,
					max = 20,
					step = 1,
					width = "full",
					order = 14,
				},
				normalFontSize = { 
					name = LATr["Achievement criterias font size"],
					type="range",
					min = 6,
					max = 20,
					step = 1,
					width = "full",
					order = 15,
				},
				maxWidth = {
					name = LATr["Achievement tracker column max width"],
					type="range",
					min = 50,
					max = 600,
					step = 5,
					width = "full",
					order = 16,
				},
				objectivescellpadding = {
					name = LATr["Achievement criterias cell padding"],
					type="range",
					min = 0,
					max = 50,
					step = 1,
					width = "full",
					order = 17,
				},
				scale = {
					name = LATr["Achievement tracker scale"],
					type = "range",
					min = 0,
					max = 1,
					step = 0.1,
					width = "full",
					order = 18,
				},
				backDropColor = {
					name = LATr["Achievement tracker background color"],
					type = "color",
					hasAlpha = true,
					width = "full",
					order = 19,
				},
				borderColor = {
					name = LATr["Achievement tracker border color"],
					type = "color",
					hasAlpha = true,
					width = "full",
					order = 20,
				},
			},
		}
	end
	return options
end

function module:OnEnable()
	anchor:SetPoint(prdb.point, prdb.x, prdb.y)
	if not prdb.showAnchor then anchor:Hide() else anchor:Show() end
	trackerAnchor = anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	if prdb.hideTrackerInCombat then
		self:RegisterEvent("PLAYER_REGEN_ENABLED","updateAchievementTracker")
		self:RegisterEvent("PLAYER_REGEN_DISABLED","updateAchievementTracker")
	end
	if interfaceVersion == 30100 then
		self:SecureHook("AddTrackedAchievement")
		self:SecureHook("RemoveTrackedAchievement")
		self:SecureHook("GetNumTrackedAchievements")
		self:RegisterEvent("TRACKED_ACHIEVEMENT_UPDATE")
	end
end

function module:OnDisable()
	anchor:Hide()
	LibQTip:Release(tracker)
	tracker = nil
	core.UnregisterAllCallbacks(self)
end

--------------
--- Events ---
--------------
function module:TRACKED_ACHIEVEMENT_UPDATE(event, aid)
	if achievements[aid] and achievements[aid].numCriterias > 0 then
		for cid = 1, achievements[aid].numCriterias do
			local cString, cType, completed, quantity, reqQuantity, charName, flags, assetID, quantityString, cid2 = GetAchievementCriteriaInfo(aid, cid)
			achievements[aid].criterias[cid] = {}
			achievements[aid].criterias[cid].name = cString
			achievements[aid].criterias[cid].cType = cType
			achievements[aid].criterias[cid].completed = completed or 0
			achievements[aid].criterias[cid].quantity = quantity
			achievements[aid].criterias[cid].reqQuantity = reqQuantity
		end
	end
end

-------------
--- Hooks ---
-------------
function module:AddTrackedAchievement(aid)
	numTrackedAchievement = numTrackedAchievement + 1
	chdb.achievementTracked[aid] = true
	achievements[aid] = {}
	_, achievements[aid].name, _, achievements[aid].completed, _, _, _, achievements[aid].description, _, achievements[aid].image = GetAchievementInfo(aid)
	achievements[aid].numCriterias = GetAchievementNumCriteria(aid)
	achievements[aid].criterias = {}
	for cid = 1, achievements[aid].numCriterias do
		local cString, cType, completed, quantity, reqQuantity, charName, flags, assetID, quantityString, cid2 = GetAchievementCriteriaInfo(aid, cid)
		achievements[aid].criterias[cid] = {}
		achievements[aid].criterias[cid].name = cString
		achievements[aid].criterias[cid].cType = cType
		achievements[aid].criterias[cid].completed = completed or 0
		achievements[aid].criterias[cid].quantity = quantity
		achievements[aid].criterias[cid].reqQuantity = reqQuantity
	end
	achievements[aid].category = GetAchievementCategory(aid)
	if not achievementsByCategory[achievements[aid].category] then
		achievementsByCategory[achievements[aid].category] = {}
		achievementsByCategory[achievements[aid].category].numAchievements = 0
		achievementsByCategory[achievements[aid].category].achievements ={}
	end
	achievementsByCategory[achievements[aid].category].numAchievements = achievementsByCategory[achievements[aid].category].numAchievements + 1
	achievementsByCategory[achievements[aid].category].achievements[aid] = true
	module:updateAchievementTracker()
end

function module:RemoveTrackedAchievement(aid)
	numTrackedAchievement = numTrackedAchievement - 1
	chdb.achievementTracked[aid] = nil
	achievements[aid].name = nil
	achievements[aid].description = nil
	achievements[aid].image = nil
	achievementsByCategory[achievements[aid].category].numAchievements = achievementsByCategory[achievements[aid].category].numAchievements - 1
	if achievementsByCategory[achievements[aid].category].numAchievements == 0 then
		achievementsByCategory[achievements[aid].category].numAchievements = nil
		achievementsByCategory[achievements[aid].category].achievements[aid] = nil
		achievementsByCategory[achievements[aid].category].achievements = nil
	end
	module:updateAchievementTracker()
end

function module:GetNumTrackedAchievements()
	return numTrackedAchievement
end

----------------------------
--- Custom Cell Provider ---
----------------------------
local cProvider, cPrototype, basecCellPrototype = LibQTip:CreateCellProvider(LibQTip.LabelProvider)
function cPrototype:InitializeCell()
	self.fontString = self:CreateFontString()
	self.fontString:SetFontObject(GameTooltipText)
end

function cPrototype:SetupCell(tooltip, value, justification, font, bType, bValue, bValue2, maxWidth, indent, anchor)
	self:EnableMouse(true)


	indent = indent or 0
	local width, height = basecCellPrototype.SetupCell(self, tooltip, value, justification, font)
	local fs = self.fontString
	fs:SetPoint("TOPLEFT", self, "TOPLEFT", indent, 0)
	fs:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", 0, 0)
	width = width + indent
	if width > maxWidth then
		fs:SetWidth(maxWidth)
		height = fs:GetHeight()
		width = maxWidth
	end
	fs:Show()
	return width, height
end

function cPrototype:ReleaseCell() 
	self:EnableMouse(false) 
	self:ClearAllPoints()
	self:SetScript("OnEnter", nil) 
	self:SetScript("OnLeave", nil) 
	self:SetScript("OnMouseDown", nil) 
end

function module:updateAchievementTracker()
	if (not prdb.toggleTracker) or (numTrackedAchievement == 0) or (prdb.hideTrackerInCombat and UnitAffectingCombat("player")) then return end
	tracker = LibQTip:Acquire("tomQuest2Achievements", 2, "LEFT", "RIGHT")
	tracker:EnableMouse(true)
	tracker:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])
	tracker:SetBackdropBorderColor(prdb.borderColor[1], prdb.borderColor[2], prdb.borderColor[3], prdb.borderColor[4])
	tracker:SetScale(prdb.scale)
	tracker:SetFrameStrata("BACKGROUND")

	local headerFont = CreateFont("headerFont")
	headerFont:SetFont(LSM:Fetch("font", prdb.font), prdb.headerFontSize)
	tracker:SetHeaderFont(headerFont)

	local smallFont = CreateFont("smallFont")
	smallFont:SetFont(LSM:Fetch("font", prdb.font), prdb.normalFontSize)
	tracker:SetFont(smallFont)

	local tempLine
	local plus =  format("|TInterface\\Buttons\\UI-PlusButton-Up:%d|t", 24)
	local minus = format("|TInterface\\Buttons\\UI-MinusButton-Up:%d|t", 24)
	for category in pairs(achievementsByCategory) do
		chdb.categoryDisplay[category] = chdb.categoryDisplay[category] or 1
		tempLine = tracker:AddLine()
		tracker:SetCell(tempLine, 1, format("%s |cffffffff%s|r", chdb.categoryDisplay[category] == 0 and plus or minus, category), headerFont, "LEFT", 2)
		if chdb.categoryDisplay[category] == 1 then
			for aid in pairs(achievementsByCategory[category].achievements) do
				tempLine = tracker:AddLine()
				tracker:SetCell(tempLine, 1, format("|T%s:24|t |cffffffff%s|r", achievements[aid].image, achievements[aid].name), headerFont, "LEFT", 2)
			--[[	for objective in pairs(core.questsObjectives[uid]) do
					if prdb.hideCompletedObjective == 3 or (prdb.hideCompletedObjective == 2 and core.questsObjectives[uid][objective].partyComplete == 0) or (prdb.hideCompletedObjective == 1 and core.questsObjectives[uid][objective].playerComplete == 0) then
						tempLine = tracker:AddLine()
						tracker:SetCell(tempLine, 1, format("|cff%s%s|r", prdb.colorizeQuestObjective and core.questsObjectives[uid][objective].qoColor or "ffffff", core.questsObjectives[uid][objective].desc), smallFont, "LEFT", 1, cProvider, "quest", uid, objective, prdb.maxWidth, prdb.objectivescellpadding, trackerAnchor)
						tracker:SetCell(tempLine, 2, core.questsObjectives[uid][objective].qoStatus, smallFont, "RIGHT", 1, cProvider, "quest", uid, objective, prdb.maxWidth, 0, trackerAnchor)
					end
				end]]
			end
		end
	end
	
	tracker:SetClampedToScreen(true)
	tracker:ClearAllPoints()
	if prdb.anchorPosition == 1 then
		if prdb.showAnchor then
			tracker:SetPoint("TOPLEFT", anchor, "BOTTOMLEFT")
		else
			tracker:SetPoint("TOPLEFT", anchor, "BOTTOMLEFT", 0, 24)
		end
	elseif prdb.anchorPosition == 2 then
		if prdb.showAnchor then
			tracker:SetPoint("TOP", anchor, "BOTTOM")
		else
			tracker:SetPoint("TOP", anchor, "BOTTOM", 0, 24)
		end
	elseif prdb.anchorPosition == 3 then
		if prdb.showAnchor then
			tracker:SetPoint("TOPRIGHT", anchor, "BOTTOMRIGHT")
		else
			tracker:SetPoint("TOPRIGHT", anchor, "BOTTOMRIGHT", 0, 24)
		end
	elseif prdb.anchorPosition == 4 then
		if prdb.showAnchor then
			tracker:SetPoint("BOTTOMLEFT", anchor, "TOPLEFT")
		else
			tracker:SetPoint("BOTTOMLEFT", anchor, "TOPLEFT", 0, -24)
		end
	elseif prdb.anchorPosition == 5 then
		if prdb.showAnchor then
			tracker:SetPoint("BOTTOM", anchor, "TOP")
		else
			tracker:SetPoint("BOTTOM", anchor, "TOP", 0, -24)
		end
	elseif prdb.anchorPosition == 6 then
		if prdb.showAnchor then
			tracker:SetPoint("BOTTOMRIGHT", anchor, "TOPRIGHT")
		else
			tracker:SetPoint("BOTTOMRIGHT", anchor, "TOPRIGHT", 0, -24)
		end
	end
	tracker:Show()
end